#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include <xs.h>
#include "xennerctrl.h"
#include <xen/io/xenbus.h>
#include <xen/io/netif.h>

#include "shared.h"
#include "list.h"
#include "daemon.h"
#include "xenbackd.h"

/* ------------------------------------------------------------- */

#define BE_ROOT "/local/domain/0/backend/netbackd"

enum netif_state {
    NETIF_UNKNOWN = 0,
    NETIF_PROBED,
    NETIF_INITIALISED,
    NETIF_CONNECTED,
    NETIF_DISCONNECTED,
};

struct netdev {
    struct xendev         xendev;  /* must be first */
    enum netif_state      state;
    char                  *mac;
    char                  *bridge;
    int                   tun;
    int                   tx_work;
    int                   tx_ring_ref;
    int                   rx_ring_ref;
    netif_tx_back_ring_t  tx_ring;
    netif_rx_back_ring_t  rx_ring;

    /* worker threads */
    pthread_t             tx_thread;
    pthread_t             rx_thread;
    pthread_cond_t        tx_waiter;
    pthread_mutex_t       tx_waitlock;
};

/* ------------------------------------------------------------- */

static void net_tx_thread_wakeup(struct netdev *netdev)
{
    if (!netdev->tx_thread)
	return;
    pthread_mutex_lock(&netdev->tx_waitlock);
    netdev->tx_work++;
    pthread_cond_signal(&netdev->tx_waiter);
    pthread_mutex_unlock(&netdev->tx_waitlock);
}

static void net_tx_thread_gosleep(struct netdev *netdev)
{
    pthread_mutex_lock(&netdev->tx_waitlock);
    if (!netdev->tx_work)
	pthread_cond_wait(&netdev->tx_waiter, &netdev->tx_waitlock);
    netdev->tx_work = 0;
    pthread_mutex_unlock(&netdev->tx_waitlock);
}
 
static void net_tx_response(struct netdev *netdev, netif_tx_request_t *txp, int8_t st)
{
    RING_IDX i = netdev->tx_ring.rsp_prod_pvt;
    netif_tx_response_t *resp;
    int notify;
    
    resp = RING_GET_RESPONSE(&netdev->tx_ring, i);
    resp->id     = txp->id;
    resp->status = st;
    
#if 0
    if (txp->flags & NETTXF_extra_info)
	RING_GET_RESPONSE(&netdev->tx_ring, ++i)->status = NETIF_RSP_NULL;
#endif
    
    netdev->tx_ring.rsp_prod_pvt = ++i;
    RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&netdev->tx_ring, notify);
    if (notify)
	xenner_evtchn_notify(evtchnd, netdev->xendev.local_port);
    
    if (i == netdev->tx_ring.req_cons) {
	int more_to_do;
	RING_FINAL_CHECK_FOR_REQUESTS(&netdev->tx_ring, more_to_do);
	if (more_to_do)
	    netdev->tx_work++;
    }
}

static void net_tx_error(struct netdev *netdev, netif_tx_request_t *txp, RING_IDX end)
{
#if 0
	RING_IDX cons = netdev->tx_ring.req_cons;

	do {
		make_tx_response(netif, txp, NETIF_RSP_ERROR);
		if (cons >= end)
			break;
		txp = RING_GET_REQUEST(&netdev->tx_ring, cons++);
	} while (1);
	netdev->tx_ring.req_cons = cons;
	netif_schedule_work(netif);
	netif_put(netif);
#else
	net_tx_response(netdev, txp, NETIF_RSP_ERROR);
#endif
}

static void *net_tx_thread(void *arg)
{
    struct netdev *netdev = arg;
    netif_tx_request_t txreq;
    RING_IDX rc, rp;
    void *page;
    int len, status;
    
    d1printf("%s: start\n", __FUNCTION__);
    for (;;) {
	if (netdev->state != NETIF_CONNECTED)
	    break;

	rc = netdev->tx_ring.req_cons;
	rp = netdev->tx_ring.sring->req_prod;
	rmb(); /* Ensure we see queued requests up to 'rp'. */
	
	while ((rc != rp)) {
	    if (RING_REQUEST_CONS_OVERFLOW(&netdev->tx_ring, rc))
		break;
	    memcpy(&txreq, RING_GET_REQUEST(&netdev->tx_ring, rc), sizeof(txreq));
	    netdev->tx_ring.req_cons = ++rc;

#if 1
	    /* should not happen in theory, we don't announce the *
	     * feature-{sg,gso,whatelse} flags in xenstore (yet?) */
	    if (txreq.flags & NETTXF_extra_info) {
		d1printf("%s: FIXME: extra info flag\n", __FUNCTION__);
		net_tx_error(netdev, &txreq, rc);
		continue;
	    }
	    if (txreq.flags & NETTXF_extra_info) {
		d1printf("%s: FIXME: more data flag\n", __FUNCTION__);
		net_tx_error(netdev, &txreq, rc);
		continue;
	    }
#endif

	    if (txreq.size < 14) {
		d1printf("%s: bad packet size: %d\n", __FUNCTION__, txreq.size);
		net_tx_error(netdev, &txreq, rc);
		continue;
	    }

	    if ((txreq.offset + txreq.size) > PAGE_SIZE) {
		d1printf("%s: error: page crossing\n", __FUNCTION__);
		net_tx_error(netdev, &txreq, rc);
		continue;
	    }

#if 0
	    d1printf("%s: packet ref %d, off %d, len %d, flags 0x%x%s%s%s%s\n",
		     __FUNCTION__, txreq.gref, txreq.offset, txreq.size, txreq.flags,
		     (txreq.flags & NETTXF_csum_blank)     ? " csum_blank"     : "",
		     (txreq.flags & NETTXF_data_validated) ? " data_validated" : "",
		     (txreq.flags & NETTXF_more_data)      ? " more_data"      : "",
		     (txreq.flags & NETTXF_extra_info)     ? " extra_info"     : "");
#endif

	    page = vm_ref2ptr(netdev->xendev.domain, txreq.gref);
	    if (NULL == page) {
		d1printf("%s: error: gref dereference failed\n", __FUNCTION__);
		net_tx_error(netdev, &txreq, rc);
		continue;
	    }
	    len = write(netdev->tun, page + txreq.offset, txreq.size);
	    status = (len == txreq.size) ? NETIF_RSP_OKAY : NETIF_RSP_ERROR;
	    net_tx_response(netdev, &txreq, status);
	}

	net_tx_thread_gosleep(netdev);
    }
    d1printf("%s: exit\n", __FUNCTION__);
    return NULL;
}

/* ------------------------------------------------------------- */

static int wait_data(int fd)
{
    fd_set rd;

    FD_ZERO(&rd);
    FD_SET(fd,&rd);
    return select(fd+1, &rd, NULL, NULL, NULL);
}

static void net_rx_response(struct netdev *netdev,
			    netif_rx_request_t *req,
			    int8_t st, uint16_t size,
			    uint16_t flags)
{
    RING_IDX i = netdev->rx_ring.rsp_prod_pvt;
    netif_rx_response_t *resp;
    int notify;
    
    resp = RING_GET_RESPONSE(&netdev->rx_ring, i);
    resp->offset     = 0;
    resp->flags      = flags;
    resp->id         = req->id;
    resp->status     = (int16_t)size;
    if (st < 0)
	resp->status = (int16_t)st;
    
    netdev->rx_ring.rsp_prod_pvt = ++i;
    RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&netdev->rx_ring, notify);
    if (notify)
	xenner_evtchn_notify(evtchnd, netdev->xendev.local_port);
}

static void *net_rx_thread(void *arg)
{
    netif_rx_request_t rxreq;
    struct netdev *netdev = arg;
    RING_IDX rc, rp;
    int len, status;
    void *page;
    
    d1printf("%s: start\n", __FUNCTION__);
    for (;;) {
	if (netdev->state != NETIF_CONNECTED)
	    break;
	if (1 != wait_data(netdev->tun))
	    continue;

	rc = netdev->rx_ring.req_cons;
	rp = netdev->rx_ring.sring->req_prod;
	rmb(); /* Ensure we see queued requests up to 'rp'. */
	
	if (rc == rp || RING_REQUEST_CONS_OVERFLOW(&netdev->rx_ring, rc)) {
	    char buf[PAGE_SIZE];
	    d1printf("%s: no buffer, drop packet\n", __FUNCTION__);
	    read(netdev->tun, buf, sizeof(buf));
	    continue;
	}

	memcpy(&rxreq, RING_GET_REQUEST(&netdev->rx_ring, rc), sizeof(rxreq));
	netdev->rx_ring.req_cons = ++rc;

	page = vm_ref2ptr(netdev->xendev.domain, rxreq.gref);
	if (NULL == page) {
	    d1printf("%s: error: gref dereference failed\n", __FUNCTION__);
	    net_rx_response(netdev, &rxreq, NETIF_RSP_ERROR, 0, 0);
	    continue;
	}

	len = read(netdev->tun, page, PAGE_SIZE);
	status = (len > 0) ? NETIF_RSP_OKAY : NETIF_RSP_ERROR;
	net_rx_response(netdev, &rxreq, status, len, 0);
    }
    d1printf("%s: exit\n", __FUNCTION__);
    return NULL;
}

/* ------------------------------------------------------------- */

static void net_setup_backend(struct netdev *netdev)
{
    struct ifreq ifreq;
    
    if (netdev->state != NETIF_PROBED)
	return;
    
    /* read xenstore entries */
    if (NULL == netdev->mac)
	netdev->mac = read_be_str(&netdev->xendev, "mac");
    if (NULL == netdev->bridge)
	netdev->bridge = read_be_str(&netdev->xendev, "bridge");

    /* do we have all we need? */
    if (NULL == netdev->mac ||
	NULL == netdev->bridge)
	return;

    /* setup tun device */
    if (-1 == netdev->tun) {
	netdev->tun = open("/dev/net/tun", O_RDWR);
	if (-1 == netdev->tun) {
	    d1printf("can't open %s: %s\n", "/dev/net/tun", strerror(errno));
	    return;
	}
	memset(&ifreq, 0, sizeof(ifreq));
	ifreq.ifr_flags = IFF_TAP | IFF_NO_PI;
	strncpy(ifreq.ifr_name, "xenner%d", IFNAMSIZ);
	if (ioctl(netdev->tun, TUNSETIFF, &ifreq) < 0) {
	    d1printf("ioctl TUNSETIFF: %s\n", strerror(errno));
	    goto err_tun;
	}
	d1printf("%s: created interface %s\n", __FUNCTION__, ifreq.ifr_name);

#if 1 /* FIXME */
	if (0 == fork())
	    execlp("brctl", "brctl", "addif", netdev->bridge, ifreq.ifr_name, NULL);
	if (0 == fork())
	    execlp("ip", "ip", "link", "set", ifreq.ifr_name, "up", NULL);
#endif
    }

    /* fill info */
    write_be_int(&netdev->xendev, "feature-rx-copy", 1);
    write_be_int(&netdev->xendev, "feature-rx-flip", 0);

    change_state_xendev(&netdev->xendev, XenbusStateInitWait);
    netdev->state = NETIF_INITIALISED;
    return;

err_tun:
    close(netdev->tun);
    netdev->tun = -1;
    return;
}

static void net_connect_frontend(struct netdev *netdev)
{
    struct netif_tx_sring *txs;
    struct netif_rx_sring *rxs;
    int rx_copy;
    
    if (netdev->state != NETIF_INITIALISED)
	return;

    netdev->tx_ring_ref = read_fe_int(&netdev->xendev, "tx-ring-ref");
    netdev->rx_ring_ref = read_fe_int(&netdev->xendev, "rx-ring-ref");
    netdev->xendev.remote_port = read_fe_int(&netdev->xendev, "event-channel");
    rx_copy = read_fe_int(&netdev->xendev, "request-rx-copy");
    if (0 == rx_copy ||
	0 == netdev->tx_ring_ref ||
	0 == netdev->rx_ring_ref ||
	0 == netdev->xendev.remote_port)
	return;

    txs = vm_ref2ptr(netdev->xendev.domain, netdev->tx_ring_ref);
    rxs = vm_ref2ptr(netdev->xendev.domain, netdev->rx_ring_ref);
    if (!txs || !rxs)
	return;
    BACK_RING_INIT(&netdev->tx_ring, txs, PAGE_SIZE);
    BACK_RING_INIT(&netdev->rx_ring, rxs, PAGE_SIZE);

    netdev->xendev.local_port = xenner_evtchn_bind_interdomain
	(evtchnd, netdev->xendev.dom, netdev->xendev.remote_port);

    d1printf("%s: ok: tx-ring-ref %d, rx-ring-ref %d, "
	     "remote port %d, local port %d\n", __FUNCTION__,
	     netdev->tx_ring_ref, netdev->rx_ring_ref,
	     netdev->xendev.remote_port, netdev->xendev.local_port);

    change_state_xendev(&netdev->xendev, XenbusStateConnected);
    netdev->state = NETIF_CONNECTED;

    pthread_create(&netdev->tx_thread, NULL, net_tx_thread, netdev);
    pthread_create(&netdev->rx_thread, NULL, net_rx_thread, netdev);
}

static void net_check_ring(struct netdev *netdev)
{
    if (netdev->state != NETIF_CONNECTED)
	return;
    net_tx_thread_wakeup(netdev);
}

static void net_disconnect(struct netdev *netdev, enum xenbus_state state)
{
    if (-1 != netdev->tun) {
	close(netdev->tun);
	netdev->tun = -1;
    }
    if (netdev->xendev.local_port) {
	xenner_evtchn_unbind(evtchnd, netdev->xendev.local_port);
	netdev->xendev.local_port = 0;
    }

    if (state)
	change_state_xendev(&netdev->xendev, state);
    netdev->state = NETIF_DISCONNECTED;

    if (netdev->tx_thread)
	net_tx_thread_wakeup(netdev);
    if (netdev->rx_thread)
	pthread_kill(netdev->rx_thread, SIGUSR2);
}

/* ------------------------------------------------------------- */

static int net_alloc(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    netdev->state = NETIF_PROBED;
    netdev->tun   = -1;
    return 0;
}

static int net_setup(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    net_setup_backend(netdev);
    net_connect_frontend(netdev);
    net_check_ring(netdev);
    return 0;
}

static int net_backend(struct xendev *xendev, char *node, char *val)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, netdev, node,
	     val ? val : "<deleted>");
    net_setup_backend(netdev);
    return 0;
}

static int net_frontend(struct xendev *xendev, char *node, char *val)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, netdev, node,
	     val ? val : "<deleted>");
    switch (xendev->fe_state) {
    case XenbusStateInitialised:
    case XenbusStateConnected:
	net_connect_frontend(netdev);
	net_check_ring(netdev);
	break;

    case XenbusStateClosing:
    case XenbusStateClosed:
	net_disconnect(netdev, xendev->fe_state);
	break;

    case XenbusStateInitialising:
    case XenbusStateInitWait:
    case XenbusStateUnknown:
	/* make gcc happy */
	break;
    }
    return 0;
}

static int net_event(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    net_check_ring(netdev);
    return 0;
}

static int net_free(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);
    void *dummy;

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    net_disconnect(netdev, 0);

    if (netdev->tx_thread)
	pthread_join(netdev->tx_thread, &dummy);
    if (netdev->rx_thread)
	pthread_join(netdev->rx_thread, &dummy);
    return 0;
}

static struct devops netdev_ops = {
    .size  = sizeof(struct netdev),
    .alloc = net_alloc,
    .setup = net_setup,
    .xs_be = net_backend,
    .xs_fe = net_frontend,
    .event = net_event,
    .free  = net_free,
};

/* ------------------------------------------------------------- */

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "netbackd --  xenner network backend daemon\n"
	    "\n"
	    "usage: netbackd [options]\n"
	    "options:\n"
	    "   -h            print this text\n"
	    "   -d            increase debuglevel\n"
	    "   -p <file>     specify pidfile\n"
	    "\n"
	    "-- \n"
	    "(c) 2007 Gerd Hoffmann <kraxel@redhat.com>\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    int c;

    for (;;) {
        if (-1 == (c = getopt(argc, argv, "hdp:")))
            break;
        switch (c) {
        case 'd':
	    debug++;
	    break;
	case 'p':
	    pidfile = optarg;
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }

    return mainloop(BE_ROOT, &netdev_ops);
}
